%  Copyright 2015 Kwang In Kim, Max-Planck-Institut for Informatics
%  Please refer to license.txt for sanctioned usage.
% 
%  The code constructs a global high-order regularization matrix based on local high-order regularization matrices:
%  Kwang In Kim, James Tompkin, Hanspeter Pfister, Christian Theobalt,
%  "Local high-order regularization on data manifolds", CVPR 2015.
%
% [L] = ConstructLGOpr(Nodes, NNIdx, SigmaSq)
% INPUT:
% 'Nodes': (N times d) data matrix; 'N': number of data points, 'd': 
%           dimensionality of data.
% 'NNIdx': (N times k) matrix containing indices of k-nearest neighbors for
%           each data point (This matrix can be obtained using GetKNN).
% 'ManDim': an integer specifying the dimensionality of the manifold
%             (for SSR we recommend to do cross-validation over this parameter if
%             the dimensionality of the manifold is unknown).
%
% OUTPUT:
% 'L': (N times N) matrix - local Gaussian regularization matrix L.

function [L] = ConstructLGOpr(Nodes, NNIdx, ManDim)

[NumNodes,AmbDim] = size(Nodes);

%%% sparse version
    BlSpMatSize = 300000;
    MaxNumNN = size(NNIdx,2);
    NumNonzero = NumNodes*MaxNumNN^2;
    BlSpMatNum = ceil(NumNonzero/BlSpMatSize);
    BlSpL = cell(BlSpMatNum,1);
%%% sparse version

SpMatIdx = 1;
SpEleCounter = 1;
SpRowIdx = zeros(BlSpMatSize,1);
SpColIdx = zeros(BlSpMatSize,1);
LVal = zeros(BlSpMatSize,1);
for i=1:NumNodes
    NumNeighbors = find(NNIdx(i,:)<=0,1);
    if ~isempty(NumNeighbors)
        NumNeighbors = NumNeighbors-1;
    else
        NumNeighbors = length(NNIdx(i,:));
    end
    G = Nodes(NNIdx(i,1:NumNeighbors),:);   

    %%% constructs the local approximations: line 3 of Algorithm 1 in the main paper
        CtBase = Nodes(i,:);
        [NG,TS] = GetNormalCoordGDim(CtBase, G, ManDim);
        [NBase,TS] = GetNormalCoordGGivenTS(CtBase,TS);
        BaseIndexInG = find(NNIdx(i,1:NumNeighbors)==i,1);
    %%% constructs the local approximations: line 3 of Algorithm 1 in the main paper
    
    NGNormSq = sum(NG.^2,2)';
    NBaseNormSq = sum(NBase.^2,2)';
    DistSq = NG*NBase';
    DistSq = DistSq';
    DistSq = abs(NGNormSq-2*DistSq+NBaseNormSq);

    RParam = mean(sqrt(DistSq))*0.1;
    RParam = RParam^2;

    %%% calculates the local regularization sub-matrices: : line 4 of Algorithm 1 in the main paper
        [LocalL] = ConstructLocalRegOperatorGaussianBZLN(NG, RParam, BaseIndexInG);
    %%% calculates the local regularization sub-matrices: : line 4 of Algorithm 1 in the main paper
    
    %%% re-arranges local regularization sub-matrices: line 6 of Algorithm 1 in the main paper
        for j=1:NumNeighbors
            GRIdx = NNIdx(i,j);

            GCIdx = NNIdx(i,1:NumNeighbors);
            SpRowIdx(SpEleCounter:SpEleCounter+NumNeighbors-1) = GRIdx;
            SpColIdx(SpEleCounter:SpEleCounter+NumNeighbors-1) = GCIdx;
            LVal(SpEleCounter:SpEleCounter+NumNeighbors-1) = LocalL(j,1:NumNeighbors);
            SpEleCounter = SpEleCounter+NumNeighbors;
        end
    %%% re-arranges local regularization sub-matrices: line 6 of Algorithm 1 in the main paper
    if SpEleCounter > BlSpMatSize-NumNeighbors-1
        NonZeroIdx = length(find(SpRowIdx>0));
        SpRowIdx = SpRowIdx(1:NonZeroIdx);
        SpColIdx = SpColIdx(1:NonZeroIdx);
        LVal = LVal(1:NonZeroIdx);
        BlSpL{SpMatIdx} = sparse(SpRowIdx,SpColIdx,LVal,NumNodes,NumNodes);
        
        SpMatIdx = SpMatIdx+1;
        SpEleCounter = 1;
        SpRowIdx = zeros(BlSpMatSize,1);
        SpColIdx = zeros(BlSpMatSize,1);
        LVal = zeros(BlSpMatSize,1);
    end
end

if SpEleCounter > 1
    NonZeroIdx = length(find(SpRowIdx>0));
    SpRowIdx = SpRowIdx(1:NonZeroIdx);
    SpColIdx = SpColIdx(1:NonZeroIdx);
    LVal = LVal(1:NonZeroIdx);
    BlSpL{SpMatIdx} = sparse(SpRowIdx,SpColIdx,LVal,NumNodes,NumNodes);        
    SpMatIdx = SpMatIdx+1;
end

L = sparse([],[],[],NumNodes,NumNodes,0);
for i=1:SpMatIdx-1
    L = L+BlSpL{i};
end

L = (L+L')/2;
