%  Copyright 2015 Kwang In Kim, Max-Planck-Institut for Informatics
%  Please refer to license.txt for sanctioned usage.
%
%  The code shows a demo of regression based on local high-order regularization:
%  Kwang In Kim, James Tompkin, Hanspeter Pfister, Christian Theobalt,
%  "Local high-order regularization on data manifolds", CVPR 2015.

%  For our local Gaussian regularizer there are 3 parameters to be tuned:
%  1. 'LGParam.KNSize' specifies the size of nearest neighbors for each data point. 
%     This parameter affects the estimation of local tangent space.
%  2. 'LGParam.ManDim' specifies the dimensionality of the manifold
%  3. 'LGParam.Lambda' is the regularization parameter.

%  For comparison, we provide a code for the graph Laplacian which has 3 parameters to be tuned:
%  1. 'LapParam.KNSize' specifies the size of nearest neighbors for each data point. 
%     This parameter affects the estimation of local tangent space.
%  2. 'LapParam.SigmaSq' is the parameter for the weight function.
%  3. 'LapParam.Lambda' is the regularization parameter.

% In general, we recommend tuning the parameters using cross-validation based on given labeled data points.
% To facilitate the demo here, we provide a set of pre-optimized parameters.


addpath './LGComp/';
addpath './LapComp/';

%%% parameters for LG
    LGParam.ManDim    = 2;
    LGParam.KNSize    = 20;
    LGParam.Lambda    = 1e-7;
%%% parameters for LG

%%% parameters for Lap
    LapParam.SigmaSq   = 10;
    LapParam.KNSize    = 20;
    LapParam.Lambda    = 1e-7;
%%% parameters for Lap

%%% display toy 2D data
    XGrid = [-1:0.02:1]';
    YGrid = [-1:0.02:1]';
    XGridSize = length(XGrid);
    YGridSize = length(YGrid);

    ZDat = zeros(YGridSize,XGridSize);
    ZDat(floor(YGridSize/2),floor(XGridSize/2)) = 10;

    ZDat(1,1) = -1;
    ZDat(1,end) = -1;
    ZDat(end,1) = -1;
    ZDat(end,end) = -1;
    NumDataPoints = YGridSize*XGridSize;

    %%% preparing label data strucutre
        Data.x = zeros(NumDataPoints,2);
        Labels.y = zeros(NumDataPoints,1);
        Labels.LFlag = zeros(size(Labels.y));
        NumDataPoints = 0;
        for i=1:YGridSize
            for j=1:XGridSize
                NumDataPoints = NumDataPoints+1;
                Data.x(NumDataPoints,:) = [YGrid(i),XGrid(j)];
                if ZDat(i,j) ~= 0
                    Labels.y(NumDataPoints,:) = ZDat(i,j);
                    Labels.LFlag(NumDataPoints,:) = 1;
                end
            end
        end
    %%% preparing label data strucutre
    save Toy2D.mat XGrid YGrid XGridSize YGridSize ZDat Data Labels;
    load Toy2D.mat XGrid YGrid XGridSize YGridSize ZDat Data Labels;
%%% display toy 2D data
    
%%% display toy 2D data
    subplot(1,3,1);
    surf(XGrid,YGrid,ZDat);
    LIdx = find(Labels.LFlag>0);
    hold on;
    plot3(Data.x(LIdx,1)',Data.x(LIdx,2)',Labels.y(LIdx),'.r','MarkerSize',20);
    axis([-1 1 -1 1 -5 10])
    hold off;
%%% display toy 2D data

%%% build nearest neighbor indices
    [NNIdx] = GetKNN(Data.x, 200);
    save NNIdxTemp.mat NNIdx;
    load NNIdxTemp.mat NNIdx;
%%% build nearest neighbor indices

%%% building LG regularizer
    LGNNIdx = NNIdx(:,1:LGParam.KNSize);
    [LGReg] = ConstructLGOpr(Data.x, LGNNIdx, LGParam.ManDim);
%%% building LG regularizer
%%% solving LG system
    [LGf,LGReg] = SemiSupLearn(Labels, LGReg, LGParam.Lambda);
%%% solving LG system

%%% building Laplacian regularizer
    LapNNIdx = NNIdx(:,1:LapParam.KNSize);
    [LapReg] = ConstructLap(Data.x, LapNNIdx, LapParam.SigmaSq);
%%% building Laplacian regularizer
%%% solving Lap system
    [Lapf,LapReg] = SemiSupLearn(Labels, LapReg, LapParam.Lambda);
%%% solving Lap system

%%% display graph Laplacian results
    subplot(1,3,2);
    LapfSurfDat = zeros(YGridSize,XGridSize);
    NumDataPoints = 0;
    for i=1:YGridSize
        for j=1:XGridSize
            NumDataPoints = NumDataPoints+1;
            LapfSurfDat(i,j) = Lapf(NumDataPoints,:);
        end
    end
    surf(XGrid,YGrid,LapfSurfDat);
    title('Lap');
    hold on;
    plot3(Data.x(LIdx,1)',Data.x(LIdx,2)',Labels.y(LIdx),'.r','MarkerSize',20);
    axis([-1 1 -1 1 -5 10])
    hold off;
%%% display graph Laplacian results

%%% display LG regularier results
    subplot(1,3,3);
    LGfSurfDat = zeros(YGridSize,XGridSize);
    NumDataPoints = 0;
    for i=1:YGridSize
        for j=1:XGridSize
            NumDataPoints = NumDataPoints+1;
            LGfSurfDat(i,j) = LGf(NumDataPoints,:);
        end
    end
    surf(XGrid,YGrid,LGfSurfDat);
    title('LG');
    hold on;
    plot3(Data.x(LIdx,1)',Data.x(LIdx,2)',Labels.y(LIdx),'.r','MarkerSize',20);
    axis([-1 1 -1 1 -5 10])
    hold off;
%%% display LG regularier results

display('done.');
